//
// Copyright (C) 2000 Institut fuer Telematik, Universitaet Karlsruhe
// Copyright (C) 2004,2011 Andras Varga
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
//

#include "UDPBasicApp.h"

#include "InterfaceTableAccess.h"
#include "IPvXAddressResolver.h"
#include "NodeOperations.h"
#include "UDPControlInfo_m.h"
#include "DecisionMaker.h"


#ifdef WITH_IPv4
#include "IRoutingTable.h"
#include "RoutingTableAccess.h"
#endif

#ifdef WITH_IPv6
#include "RoutingTable6.h"
#include "RoutingTable6Access.h"
#endif


Define_Module(UDPBasicApp);

simsignal_t UDPBasicApp::sentPkSignal = registerSignal("sentPk");
simsignal_t UDPBasicApp::rcvdPkSignal = registerSignal("rcvdPk");

UDPBasicApp::UDPBasicApp() {
    selfMsg = NULL;
}

UDPBasicApp::~UDPBasicApp() {
    cancelAndDelete(selfMsg);
}

void UDPBasicApp::initialize(int stage) {
    ApplicationBase::initialize(stage);



    if (stage == 0) {

        numSent = 0;
        numReceived = 0;
        WATCH(numSent);
        WATCH(numReceived);

        localPort = par("localPort");
        destPort = par("destPort");
        startTime = par("startTime").doubleValue();
        stopTime = par("stopTime").doubleValue();


//        socket.setOutputGate(gate("udpOut"));
//        socket.bind(localPort);




        if (stopTime >= SIMTIME_ZERO && stopTime < startTime)
            error("Invalid startTime/stopTime parameters");
        selfMsg = new cMessage("sendTimer");

    }
}

void UDPBasicApp::finish() {
    recordScalar("packets BasicAPP sent", numSent);
    recordScalar("packets BasicAPP received", numReceived);
    ApplicationBase::finish();
}


void UDPBasicApp::setSocketOptions() {
    int timeToLive = par("timeToLive");
    if (timeToLive != -1)
        socket.setTimeToLive(timeToLive);

    int typeOfService = par("typeOfService");
    if (typeOfService != -1)
        socket.setTypeOfService(typeOfService);

    const char *multicastInterface = par("multicastInterface");
    if (multicastInterface[0]) {
        IInterfaceTable *ift = InterfaceTableAccess().get(this);
        InterfaceEntry *ie = ift->getInterfaceByName(multicastInterface);
        if (!ie)
            throw cRuntimeError(
                    "Wrong multicastInterface setting: no interface named \"%s\"",
                    multicastInterface);
        socket.setMulticastOutputInterface(ie->getInterfaceId());
    }

    bool receiveBroadcast = par("receiveBroadcast");
    if (receiveBroadcast)
        socket.setBroadcast(true);

    bool joinLocalMulticastGroups = par("joinLocalMulticastGroups");
    if (joinLocalMulticastGroups)
        socket.joinLocalMulticastGroups();
}

IPvXAddress UDPBasicApp::chooseDestAddr() {
    int k = intrand(destAddresses.size());
    //std::cout << k << endl;
    //int k = intrand(destAdressesRange);
    if (destAddresses.size() != 0)
        if (destAddresses[k].isIPv6() && destAddresses[k].get6().isLinkLocal()) // KLUDGE for IPv6
                {
            const char *destAddrs = par("destAddresses");
            //const char *destAddrs = check_and_cast<DecisionMaker*>(getParentModule()->getParentModule()->getSubmodule("decisionMaker"))->getDestAddress();
            cStringTokenizer tokenizer(destAddrs);
            const char *token;

            for (int i = 0; i <= k; ++i)
                token = tokenizer.nextToken();
            destAddresses[k] = IPvXAddressResolver().resolve(token);
        }


    //std::cout << getParentModule()->getFullName() << endl;
    //std::cout << getParentModule()->getIndex() << endl;
    //std::cout << getParentModule()->getFullPath() << endl;
    const char *destAddrs =
            check_and_cast<DecisionMaker*>(getParentModule()->getParentModule()->getSubmodule("decisionMaker"))->getDestAddress(getParentModule()->getFullName()).c_str();
    //std::cout << destAddrs << endl;
    //std::cout << IPvXAddressResolver().resolve(destAddrs) << endl;
    //std::cout << destAddresses[k] << endl;
    return IPvXAddressResolver().resolve(destAddrs);
}

void UDPBasicApp::sendPacket() {
    Enter_Method_Silent();
    char msgName[32];
    sprintf(msgName, "UDPBasicAppData-%d", numSent);


    cPacket *payload = new cPacket(msgName);
    payload->setByteLength(par("messageLength").longValue());

//    IPvXAddress destAddr = chooseDestAddr();

    IPvXAddress destAddr = IPv4Address::ALLONES_ADDRESS;

    emit(sentPkSignal, payload);
    socket.sendTo(payload, destAddr, destPort);
    numSent++;
}





void UDPBasicApp::processStart() {

    std::cout<<" UDPBasicApp::processStart()   " <<endl;
    std::cout<<" UDPBasicApp::processStart()   " <<endl;
    std::cout<<" UDPBasicApp::processStart()   " <<endl;

 socket.setOutputGate(gate("udpOut"));
socket.bind(localPort);
    setSocketOptions();

//  const char *localAddress = par("localAddress");
// socket.bind(*localAddress ? IPvXAddressResolver().resolve(localAddress) : IPvXAddress(), localPort);



    const char *destAddrs = par("destAddresses");
    cStringTokenizer tokenizer(destAddrs);
    const char *token;

    while ((token = tokenizer.nextToken()) != NULL) {
        IPvXAddress result;
        IPvXAddressResolver().tryResolve(token, result);
        if (result.isUnspecified())
            EV << "cannot resolve destination address: " << token << endl;
        else
            destAddresses.push_back(result);
    }


//////////////////////////////////////////////////////////////////////////////////////////////
    if (!destAddresses.empty()) {
        selfMsg->setKind(SEND);
        processSend();
    } else {
        if (stopTime >= SIMTIME_ZERO) {
            selfMsg->setKind(STOP);
            scheduleAt(stopTime, selfMsg);
        }
    }
}

void UDPBasicApp::processSend() {
   sendPacket();
    simtime_t d = simTime() + par("sendInterval").doubleValue();
    if (stopTime < SIMTIME_ZERO || d < stopTime) {
        selfMsg->setKind(SEND);
        scheduleAt(d, selfMsg);
    } else {
        selfMsg->setKind(STOP);
        scheduleAt(stopTime, selfMsg);
    }
}

void UDPBasicApp::processStop() {
    socket.close();
}


void UDPBasicApp::handleMessageWhenUp(cMessage *msg) {

    std::cout<<" UDPBasicApp::handleMessageWhenUp          "<<endl;
       std::cout<<" UDPBasicApp::handleMessageWhenUp        "  <<endl;
       std::cout<<" UDPBasicApp::handleMessageWhenUp           "  <<endl;

       std::cout<<" selfMsg->getKind()        " <<selfMsg->getKind() <<endl;
       std::cout<<" selfMsg->getKind()        " <<selfMsg->getKind() <<endl;
       std::cout<<" selfMsg->getKind()        " <<selfMsg->getKind() <<endl;


    if (msg->isSelfMessage()) {
//        ASSERT(msg == selfMsg);
        switch (selfMsg->getKind()) {
        case START:
           processStart();
            break;
        case SEND:
            processSend();
            break;
        case STOP:
            processStop();
            break;
        default:
            throw cRuntimeError("Invalid kind %d in self message",
                    (int) selfMsg->getKind());
        }
    } else if (msg->getKind() == UDP_I_DATA) {
        // process incoming packet
        processPacket(PK(msg));
    } else if (msg->getKind() == UDP_I_ERROR) {
        EV << "Ignoring UDP error report\n";
        delete msg;
    } else {
        error("Unrecognized message (%s)%s", msg->getClassName(),
                msg->getName());
    }

    if (ev.isGUI()) {
        char buf[40];
        sprintf(buf, "rcvd: %d pks\nsent: %d pks", numReceived, numSent);
        getDisplayString().setTagArg("t", 0, buf);
    }

    std::cout<<" UDPBasicApp::handleMessageWhenUp        over   "<<endl;
         std::cout<<" UDPBasicApp::handleMessageWhenUp     over   "  <<endl;
         std::cout<<" UDPBasicApp::handleMessageWhenUp     over      "  <<endl;
}

void UDPBasicApp::processPacket(cPacket *pk) {
    emit(rcvdPkSignal, pk);
    EV << "Received packet: " << UDPSocket::getReceivedPacketInfo(pk) << endl;
    delete pk;
    numReceived++;
}



bool UDPBasicApp::handleNodeStart(IDoneCallback *doneCallback) {
    simtime_t start = std::max(startTime, simTime());
    if ((stopTime < SIMTIME_ZERO) || (start < stopTime)
            || (start == stopTime && startTime == stopTime)) {
        selfMsg->setKind(START);
        scheduleAt(start, selfMsg);
    }
    return true;
}

bool UDPBasicApp::handleNodeShutdown(IDoneCallback *doneCallback) {
    if (selfMsg)
        cancelEvent(selfMsg);
    //TODO if(socket.isOpened()) socket.close();
    return true;
}

void UDPBasicApp::handleNodeCrash() {
    if (selfMsg)
        cancelEvent(selfMsg);
}

